package at.krixec.rosary

import android.annotation.SuppressLint
import androidx.compose.animation.core.Animatable
import androidx.compose.animation.core.AnimationSpec
import androidx.compose.animation.core.EaseInOutCubic
import androidx.compose.animation.core.LinearEasing
import androidx.compose.animation.core.repeatable
import androidx.compose.animation.core.tween
import androidx.compose.foundation.Canvas
import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.detectDragGestures
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.offset
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.material3.Button
import androidx.compose.material3.ButtonDefaults
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.derivedStateOf
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableFloatStateOf
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.Path
import androidx.compose.ui.graphics.PathMeasure
import androidx.compose.ui.graphics.Shadow
import androidx.compose.ui.graphics.asAndroidPath
import androidx.compose.ui.graphics.drawscope.Stroke
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.platform.LocalView
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.drawText
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.rememberTextMeasurer
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import at.krixec.rosary.ui.theme.Bead
import at.krixec.rosary.ui.theme.Cross
import at.krixec.rosary.ui.theme.SelectedBead
import kotlinx.coroutines.launch


data class RosaryPosition( var decade: Int, var bead: Int, var prayer: Int, var current: Int )

@SuppressLint( "CoroutineCreationDuringComposition" )
@Composable
fun Rosary(
    screenWidth: Dp,
    screenHeight: Dp,
    autoPray: AutoPray,
    rosary: RosaryPosition,
    fatima: Boolean,
    regina: Boolean
) {

    val density = LocalDensity.current.density
    val screenWidthInPx  = screenWidth.value * density
    val screenHeightInPx = screenHeight.value * density

    val beadSize = screenWidthInPx / 24f
    val knotSize = beadSize / 2
    var fontSizePrayer by remember { mutableStateOf( 18.sp ) }

    val path = remember { Path() }
    val autoPrayPath = remember { Path() }

    var prayer by remember { mutableIntStateOf( rosary.prayer ) }
    var decade by remember { mutableIntStateOf( rosary.decade ) }
    var bead   by remember { mutableIntStateOf( rosary.bead ) }
    // left -> 0, middle -> 1, right -> 2
    var current by remember { mutableIntStateOf( rosary.current ) }

    // I've bound the actual decade to an animation, little workaround
    var bonusDecade by remember { mutableIntStateOf( 0 ) }

    var running by remember { mutableStateOf( false ) }
    var next by remember { mutableStateOf( false ) }
    val scope = rememberCoroutineScope()
    val currentView = LocalView.current
    val textMeasurer = rememberTextMeasurer()
    var timeLeft by remember { mutableIntStateOf( 0 ) }
    val autoPrayProgress = remember { Animatable( if ( autoPray.active() ) 1f else -0.5f ) }

    if ( autoPray.needsAnimate() ) {
        autoPray.noAnimate()
        scope.launch {
            autoPrayProgress.animateTo(
                if ( autoPray.active() ) 1f else -0.5f,
                animationSpec = repeatable(
                    iterations = 1,
                    animation = tween(
                        durationMillis = 1200,
                        easing = EaseInOutCubic
                    )
                )
            )
        }
    }

    fun newTimer( id: Int ): AutoPrayTimer {

        return object: AutoPrayTimer(
            autoPray.prayerTime( rosary.prayer ) * 1000,
            1000L,
            id
        ) {
            override fun onTick( millisUntilFinished: Long ) {

                if ( iD != autoPray.timerID() )
                    cancel()
                else
                    timeLeft = ( millisUntilFinished / 1000 ).toInt()
            }

            override fun onFinish() {
                if ( autoPray.active() && iD == autoPray.timerID() )
                    next = true
            }
        }
    }

    var autoPrayTimer: AutoPrayTimer

    if ( !running && autoPray.active() && !autoPray.stopped() ) {
        running = true
        autoPrayTimer = newTimer( autoPray.timerID() )
        autoPrayTimer.start()
    }

    @Composable
    fun RosaryLoop() {

        var totalDrag by remember { mutableFloatStateOf( 0f ) }
        var hasMoved  = false
        var animating = false

        fun getTargetValue( right: Boolean ): Float {
            return if ( current == 1 )
                if ( right )
                    -0.8f
                else
                    1.335f
            else
                0.2f
        }

        var fullyShown   = remember { false }
        val pathMeasure  = remember { PathMeasure() }
        val progress     = remember { Animatable( if ( current == 1 ) 0.2f else 1.335f ) }
        val progress2    = remember { Animatable( if ( decade == 0 ) 0.2f else -1f ) }
        val animatedPath = remember {
            derivedStateOf {
                val destination = Path()
                pathMeasure.setPath( path, false )

                pathMeasure.getSegment(
                    0f,
                    pathMeasure.length + progress2.value * pathMeasure.length - pathMeasure.length * 0.2f,
                    destination
                )
                destination
            }
        }

        suspend fun moveRosary(
            to: Float,
            duration: Int,
            first: Boolean = false
        ) {

            val spec: AnimationSpec< Float > = repeatable(
                iterations = 1,
                animation = tween(
                    durationMillis = duration,
                    easing = LinearEasing
                )
            )
            if ( first )
                progress.animateTo( to, animationSpec = spec )
            else
                progress2.animateTo( to, animationSpec = spec )

        }

        fun goRight() {

            if ( !animating ) {

                if ( rosary.decade == 0 ) {

                    if ( rosary.bead < 7 ) {
                        rosary.bead += 1
                        rosary.prayer = when ( rosary.bead ) {
                            1 -> 4
                            2 -> 1
                            6 -> 2
                            7 -> 1

                            else -> -1
                        }
                        bead = rosary.bead
                        prayer = rosary.prayer
                    } else {

                        scope.launch {
                            animating = true
                            rosary.prayer = -1
                            prayer = rosary.prayer
                            bonusDecade = 1
                            moveRosary( -1f, 1500 )

                            bonusDecade = 0
                            rosary.decade += 1
                            rosary.bead   = 0
                            decade = rosary.decade
                            bead = 0
                            progress2.snapTo( -0.5f )
                            moveRosary( -1f, 1000 )

                            fullyShown = true
                            animating = false
                        }
                    }
                } else {
                    if ( rosary.decade == 5 && rosary.bead == 9 ) {

                        rosary.prayer = 2
                        rosary.bead += 1
                        prayer = 2
                        bead = rosary.bead

                    } else if ( rosary.decade == 5 && rosary.bead == 10 && fatima ) {

                        rosary.prayer = 3
                        rosary.bead += 1
                        prayer = 3
                        bead = rosary.bead

                    } else if ( rosary.decade == 5 && rosary.bead == ( if ( fatima ) 11 else 10 ) && regina ) {

                        rosary.prayer = 5
                        rosary.bead += if ( fatima ) 1 else 2
                        prayer = 5
                        bead = rosary.bead

                    } else if ( ( rosary.decade == 5 && rosary.bead == 12 ) ||
                        ( !regina && rosary.decade == 5 && rosary.bead == ( if ( fatima ) 11 else 10 ) ) ) {

                        autoPray.stop()
                        currentView.keepScreenOn = false
                        scope.launch {

                            animating = true
                            fullyShown = false
                            rosary.prayer = 0
                            prayer = 0
                            moveRosary( 0.2f, 1500 )

                            rosary.bead   = 0
                            rosary.decade = 0
                            bead = 0
                            decade = 0
                            progress2.snapTo( -1f )
                            moveRosary( 0.2f, 1000 )

                            animating = false
                        }
                    } else if ( ( current == 1 && rosary.bead >= 10 )
                        || ( current != 1 && rosary.bead >= 4 )
                    ) {

                        scope.launch {

                            animating = true
                            if ( current != 1 ) {
                                progress.snapTo( 1.335f )
                                rosary.bead = 0
                                rosary.prayer = -1
                            } else {
                                rosary.bead = if ( fatima ) 3 else 4
                                rosary.prayer = if ( fatima ) 3 else 1
                            }
                            bead = rosary.bead
                            prayer = rosary.prayer

                            if ( current == 1 )
                                if ( rosary.decade < 5 )
                                    rosary.decade += 1
                                else
                                    rosary.decade = 0

                            decade = rosary.decade

                            if ( rosary.decade != 0 ) {
                                moveRosary( getTargetValue( true ), 2000, true )

                                if ( current < 2 )
                                    current += 1
                                else
                                    current = 1

                                rosary.current = current
                            } else {
                                fullyShown = false
                                rosary.bead = 0
                                bead = 0
                                moveRosary( 0.2f, 1500 )
                            }
                            animating = false
                        }
                    } else {
                        rosary.bead += 1
                        if ( rosary.bead == 10 )
                            rosary.prayer = 2
                        if ( current != 1 )
                            rosary.prayer = 1

                        bead = rosary.bead
                        prayer = rosary.prayer
                    }
                }
            }
        }

        fun goLeft() {
            if ( rosary.decade == 0 && rosary.bead == 0 ) {
                scope.launch {

                    animating = true
                    rosary.prayer = if ( regina ) 5 else if ( fatima ) 3 else 2
                    prayer = if ( regina ) 5 else if ( fatima ) 3 else 2
                    bonusDecade = 5
                    moveRosary( -1f, 1500 )

                    bonusDecade = 0
                    rosary.decade = 5
                    rosary.bead   = if ( regina ) 12 else if ( fatima ) 11 else 10
                    decade = 5
                    bead = if ( regina ) 12 else if ( fatima ) 11 else 10
                    progress2.snapTo( -0.5f )
                    moveRosary( -1f, 1000 )

                    animating  = false
                    fullyShown = true
                }
            } else if ( rosary.decade == 1 && rosary.bead == 0 ) {

                scope.launch {

                    animating  = true
                    fullyShown = false
                    rosary.prayer = 1
                    prayer = 1
                    moveRosary( 0.2f, 1500 )

                    rosary.decade = 0
                    rosary.bead   = 7
                    decade = 0
                    bead = 7
                    progress2.snapTo( -1f )
                    moveRosary( 0.2f, 1000 )

                    animating = false
                }

            } else if ( ( current == 1 && rosary.bead <= 0 )
                || ( current != 1 && rosary.bead <= if ( fatima ) 3 else 4 )
            ) {
                scope.launch {
                    animating = true
                    if ( current != 1 ) {
                        progress.snapTo( -0.8f )
                        rosary.bead = 10
                        rosary.prayer = 2
                    } else {
                        rosary.bead = 4
                        rosary.prayer = 1
                    }

                    if ( current == 1 )
                        if ( rosary.decade > 0 )
                            rosary.decade -= 1
                        else {
                            rosary.decade = 5
                        }

                    prayer = rosary.prayer
                    decade = rosary.decade
                    bead = rosary.bead

                    moveRosary( getTargetValue( false ), 2000, true )

                    if ( current > 0 )
                        current -= 1
                    else
                        current = 1

                    rosary.current = current
                    animating = false
                }
            } else {
                rosary.bead -= if ( !( bead == 12 && !fatima ) ) 1 else 2
                bead = rosary.bead
                if ( rosary.decade != 0 )
                    rosary.prayer = -1
                else {
                    if ( rosary.bead < 7 ) {
                        rosary.prayer = when ( rosary.bead ) {
                            0 -> 0
                            1 -> 4
                            2 -> 1
                            6 -> 2
                            //7 -> 1

                            else -> -1
                        }
                    }
                }
                if ( rosary.decade == 5 && rosary.bead == 11 )
                    rosary.prayer = 3
                else if ( rosary.decade == 5 && rosary.bead == 10 )
                    rosary.prayer = 2
                else if ( current != 1 && rosary.decade != 0 )
                    rosary.prayer = 3
                prayer = rosary.prayer
            }
        }

        // invisible box to catch touch input
        Box(
            modifier = Modifier
                .fillMaxSize()
                .pointerInput( Unit ) {

                    detectTapGestures {

                        if ( autoPray.active() && !autoPray.stopped() ) {
                            next = true

                        } else {
                            goRight()
                        }
                    }
                }
                .pointerInput( Unit ) {
                    detectDragGestures(
                        onDragStart = {
                            totalDrag = 0f
                            hasMoved = false
                        }
                    ) { change, dragAmount ->
                        change.consume()

                        totalDrag += dragAmount.x

                        if ( !hasMoved && !animating && totalDrag > 40 ) {

                            hasMoved = true
                            if ( autoPray.active() && !autoPray.stopped() ) {
                                next = true

                            } else {
                                goRight()
                            }
                        } else if ( !hasMoved && !animating && totalDrag < -40 ) {

                            hasMoved = true

                            if ( autoPray.active() && !autoPray.stopped() ) {
                                autoPray.inc()
                                goLeft()
                                autoPrayTimer = newTimer( autoPray.timerID() )
                                autoPrayTimer.start()
                            } else
                                goLeft()

                        }
                    }
                }
        )

        Canvas(
            modifier = Modifier
                .background( Color.Transparent )
                .fillMaxSize()
        ) {

            if ( rosary.decade != 0 ) {

                val x0 = -( screenWidthInPx / 20f )
                val y0 = screenHeightInPx / 1.2f
                val y1 = screenHeightInPx / 1.6f
                val y2 = screenHeightInPx / 0.9f
                val y3 = screenHeightInPx / 1.2f

                path.reset()
                path.moveTo( x0, if ( fullyShown ) y0 else y0 + screenHeightInPx * ( progress2.value + 1f ) )
                path.cubicTo(
                    x1 = screenWidthInPx / 2f,
                    y1 = if ( fullyShown ) y1 else y1 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                    x2 = screenWidthInPx / 2f,
                    y2 = if ( fullyShown ) y2 else y2 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                    x3 = screenWidthInPx + screenWidthInPx / 20f,
                    y3 = if ( fullyShown ) y3 else y3 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                )

                drawPath(
                    color = Color.Gray,
                    path  = path,
                    style = Stroke(
                        width = knotSize
                    )
                )

                // side pater nosters
                for ( i in 0..5 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )
                    val mod = if ( i <= 2 ) -1 else 1

                    android.graphics.PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + mod + ( i + 2 ) * 0.045f ),
                            pos,
                            tan
                        )
                    }

                    val color = if ( rosary.prayer == 1 ) SelectedBead else Bead
                    if ( i % 2 != 0 && i < 3 ) {
                        drawCircle(
                            radius = beadSize,
                            color  = color,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else if ( i % 2 == 0 && i < 3 ) {
                        drawCircle(
                            radius = knotSize,
                            color  = Color.Gray,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else if ( i % 2 != 0 ) {
                        drawCircle(
                            radius = knotSize,
                            color  = Color.Gray,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else {
                        drawCircle(
                            radius = beadSize,
                            color  = color,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    }
                }

                // grey knots
                for ( i in 0..1 ) {

                    val mod = if ( i == 0 ) -2f else 14.5f
                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    android.graphics.PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + mod * 0.048f ),
                            pos,
                            tan
                        )
                    }

                    drawCircle(
                        radius = knotSize,
                        color  = Color.Gray,
                        center = Offset( pos[ 0 ], pos[ 1 ] )
                    )

                }
                for ( i in 0..9 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    android.graphics.PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + ( i - 0.5f ) * 0.0755f ),
                            pos,
                            tan
                        )
                    }
                    val color = if ( bead == i && prayer == -1 ) SelectedBead else Bead
                    drawCircle(
                        radius = beadSize,
                        color = color,
                        center = Offset( pos[ 0 ], pos[ 1 ] )
                    )
                }
            } else {

                // You should be able to adjust this
                // in the settings at some point
                val yShifter = 100f

                val x0 = screenWidthInPx + screenWidthInPx / 5f
                val y0 = screenHeightInPx / 1.8f + yShifter

                path.reset()
                path.moveTo( x0, y0 )
                path.cubicTo(
                    x1 = screenWidthInPx  / 2f,
                    y1 = screenHeightInPx / 1.6f + yShifter,
                    x2 = screenWidthInPx  / 1.4f,
                    y2 = screenHeightInPx / 1.4f + yShifter,
                    x3 = screenWidthInPx  / 1.6f,
                    y3 = screenHeightInPx / 1.3f + yShifter
                )

                drawPath(
                    color = Color.Gray,
                    path = animatedPath.value,
                    style = Stroke(
                        width = knotSize
                    )
                )

                for ( i in 0..5 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    when ( i ) {
                        0 -> {

                            android.graphics.PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan(
                                    length * ( progress2.value + 0.8f ),
                                    pos,
                                    tan
                                )
                            }

                            val color = Cross
                            val girth = beadSize * 1.5f
                            drawLine(
                                color = color,
                                start = Offset( pos[ 0 ] - 45.dp.toPx() - girth / 2, pos[ 1 ] ),
                                end   = Offset( pos[ 0 ] + 37.5.dp.toPx() - girth / 2, pos[ 1 ] + 48.7.dp.toPx() ),
                                strokeWidth = girth
                            )
                            drawLine(
                                color = color,
                                start = Offset( pos[ 0 ], pos[ 1 ] - 3.75.dp.toPx() ),
                                end   = Offset( pos[ 0 ] - 67.4.dp.toPx(), pos[ 1 ] + 112.36.dp.toPx() ),
                                strokeWidth = girth
                            )
                        }
                        1 -> {
                            // if the following numbers seem arbitrary, that's because they are
                            for ( j in -1..1 ) {

                                android.graphics.PathMeasure().apply {
                                    setPath( path.asAndroidPath(), false )
                                    getPosTan( length * ( progress2.value + j * 0.076f + 0.13f ),
                                        pos,
                                        tan
                                    )
                                }

                                val color = if ( rosary.bead == 7 ) SelectedBead else Bead
                                if ( j != 0 )
                                    drawCircle(
                                        radius = knotSize,
                                        color  = Color.Gray,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                                else
                                    drawCircle(
                                        radius = beadSize,
                                        color  = color,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                            }
                            android.graphics.PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan( length * ( progress2.value + 0.01f ),
                                    pos,
                                    tan
                                )
                            }

                        }
                        in 2..4 -> {
                            android.graphics.PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan( length * ( progress2.value + i * 0.105f + 0.08f ),
                                    pos,
                                    tan
                                )
                            }

                            val color = if ( rosary.bead == 7 - i ) SelectedBead else Bead
                            drawCircle(
                                radius = beadSize,
                                color  = color,
                                center = Offset(pos[ 0 ], pos[ 1 ])
                            )
                        }
                        else -> {
                            for ( j in -1..1 ) {

                                android.graphics.PathMeasure().apply {
                                    setPath( path.asAndroidPath(), false )
                                    getPosTan(
                                        length * ( progress2.value + i * 0.125f + j * 0.077f + 0.035f ),
                                        pos,
                                        tan
                                    )
                                }

                                val color = if ( rosary.bead == 2 ) SelectedBead else Bead
                                if ( j != 0 )
                                    drawCircle(
                                        radius = knotSize,
                                        color  = Color.Gray,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                                else
                                    drawCircle(
                                        radius = beadSize,
                                        color  = color,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                            }
                        }
                    }
                }
            }
        }

        Text(
            getPrayer( prayer, decade, bonusDecade, bead ),
            modifier = Modifier
                .fillMaxHeight( 1 / 1.4f )
                .fillMaxWidth()
                .padding( 20.dp, screenHeight / 9 ),
            textAlign = TextAlign.Center,
            color = Color.LightGray,
            fontSize = fontSizePrayer,
            onTextLayout = {
                if ( it.didOverflowHeight ) {
                    fontSizePrayer *= .95f
                }
            },
            fontWeight = FontWeight.ExtraBold,
            style = TextStyle.Default.copy(
                shadow = Shadow( color = Color.Black, offset = Offset( 4f, 4f ) )
            )
        )

        if ( next ) {
            next = false
            autoPray.inc()
            goRight()

            if ( !autoPray.stopped() ) {

                autoPrayTimer = newTimer( autoPray.timerID() )
                autoPrayTimer.start()
            }
        }
    }

    @Composable
    fun AutoPrayButton() {

        var symbol by remember { mutableStateOf( "▶" ) }
        symbol = getSymbol( autoPray.stopped() )

        Box(
            modifier = Modifier
                .fillMaxHeight()
        ) {

            Button(
                modifier = Modifier
                    .size( 100.dp )
                    .offset( 0.dp, ( -10 ).dp)
                    .align( Alignment.BottomStart ),
                colors = ButtonDefaults.buttonColors(
                    containerColor = Color.Transparent,
                    disabledContainerColor = Color.Transparent
                ),
                enabled = autoPray.active(),
                onClick = {
                    if ( autoPray.stopped() ) {
                        currentView.keepScreenOn = true
                        symbol = "■"
                        autoPray.inc()
                        autoPrayTimer = newTimer( autoPray.timerID() )
                        autoPrayTimer.start()
                        autoPray.pray()
                    } else {
                        currentView.keepScreenOn = false
                        symbol = "▶"
                        autoPray.stop()
                        autoPray.inc()
                    }
                }
            ) { }

            Canvas(
                modifier = Modifier
                    .size( screenWidth / 3f, screenHeight / 4f )
                    .align( Alignment.BottomStart )
            ) {
                val fixedScreenWidthInPx  = screenWidthInPx / 3f
                val fixedScreenHeightInPx = screenHeightInPx / 4f

                val x0 = -10f
                val y0 = fixedScreenHeightInPx / 2.2f
                val x1 = fixedScreenWidthInPx  / 1.8f
                val y1 = fixedScreenHeightInPx / 1.5f
                val x2 = -20f
                val y2 = fixedScreenHeightInPx / 1.2f
                val x3 = fixedScreenWidthInPx / 2.5f


                autoPrayPath.reset()
                autoPrayPath.moveTo( x0, y0 )
                autoPrayPath.cubicTo(
                    x1 = x1 * autoPrayProgress.value,
                    y1 = y1,
                    x2 = x2 * autoPrayProgress.value,
                    y2 = y2,
                    x3 = x3 * autoPrayProgress.value,
                    y3 = fixedScreenHeightInPx + 10f
                )

                drawPath(
                    color = Color.Gray,
                    path = autoPrayPath,
                    style = Stroke(
                        width = knotSize
                    )
                )

                val style  = TextStyle.Default.copy(
                    color = Color.LightGray,
                    fontSize = 28.sp,
                    shadow = Shadow( color = Color.Black, offset = Offset( 4f, 4f ) )
                )
                val numbersStyle = TextStyle.Default.copy(
                    color = Color.Gray,
                    fontSize = 20.sp
                )
                val textLayoutResult = textMeasurer.measure( symbol, style )

                for ( i in -1..1 ) {
                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    android.graphics.PathMeasure().apply {
                        setPath( autoPrayPath.asAndroidPath(), false )
                        getPosTan(
                            length * ( 0.48f + 0.185f * i ),
                            pos,
                            tan
                        )
                    }

                    if ( i == 0 )
                        drawCircle(
                            radius  = beadSize,
                            color   = if ( autoPrayProgress.value == -0.5f) Color.Transparent else Bead,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    else
                        drawCircle(
                            radius = knotSize,
                            color  = Color.Gray,
                            center  = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                }

                val pos = FloatArray( 2 )
                val tan = FloatArray( 2 )

                android.graphics.PathMeasure().apply {
                    setPath( autoPrayPath.asAndroidPath(), false )
                    getPosTan(
                        length * 0.5f,
                        pos,
                        tan
                    )
                }

                drawText(
                    textMeasurer = textMeasurer,
                    text    = symbol,
                    style   = style,
                    topLeft = Offset(
                        pos[ 0 ] - textLayoutResult.size.width / 2f + knotSize / 25f,
                        pos[ 1 ] - textLayoutResult.size.height / 2f - knotSize / 1.7f
                    )
                )

                if ( !autoPray.stopped() )
                    drawText(
                        textMeasurer = textMeasurer,
                        text = ( timeLeft + 1 ).toString(),
                        style = numbersStyle,
                        topLeft = Offset(
                            pos[ 0 ] + textLayoutResult.size.width,
                            pos[ 1 ] - textLayoutResult.size.height
                        )
                    )
            }
        }
    }

    RosaryLoop()
    AutoPrayButton()
}
